from typing import TypedDict, Optional, List, Dict, Literal, Optional


TrackQuality = Literal["LOW", "HIGH", "LOSSLESS", "HI_RES_LOSSLESS"]
ManifestMimeType = Literal["application/dash+xml", "application/vnd.tidal.bts"]


class TrackStream(TypedDict):
    trackId: int
    assetPresentation: Literal["FULL"]
    audioMode: Literal["STEREO"]
    audioQuality: TrackQuality
    manifestMimeType: ManifestMimeType
    manifestHash: str
    manifest: str
    albumReplayGain: float
    albumPeakAmplitude: float
    trackReplayGain: float
    trackPeakAmplitude: float
    bitDepth: Optional[int]
    sampleRate: Optional[int]


class _Artist(TypedDict):
    id: int
    name: str
    type: str
    picture: Optional[str]


class _Album(TypedDict):
    id: int
    title: str
    cover: str
    vibrantColor: str
    videoCover: Optional[str]


class Track(TypedDict):
    id: int
    title: str
    duration: int
    replayGain: float
    peak: float
    allowStreaming: bool
    streamReady: bool
    adSupportedStreamReady: bool
    djReady: bool
    stemReady: bool
    streamStartDate: str
    premiumStreamingOnly: bool
    trackNumber: int
    volumeNumber: int
    version: Optional[str]
    popularity: int
    copyright: str
    bpm: Optional[int]
    url: str
    isrc: str
    editable: bool
    explicit: bool
    audioQuality: str
    audioModes: List[str]
    mediaMetadata: Dict[str, List[str]]
    artist: _Artist
    artists: List[_Artist]
    album: _Album
    mixes: Dict[str, str]



import logging
import requests
import json

from os import makedirs
from xml.etree.ElementTree import fromstring
from base64 import b64decode
from typing import TypedDict, List

from .types.track import ManifestMimeType


logger = logging.getLogger("download")


def decodeManifest(manifest: str):
    return b64decode(manifest).decode()


def parseManifest(manifest: str):
    class AudioFileInfo(TypedDict):
        mimeType: str
        codecs: str
        encryptionType: str
        urls: List[str]

    data: AudioFileInfo = json.loads(manifest)
    return data


def parseManifestXML(xml_content: str):
    """
    Parses XML manifest file of the track.
    """

    NS = "{urn:mpeg:dash:schema:mpd:2011}"

    tree = fromstring(xml_content)

    representationElement = tree.find(
        f"{NS}Period/{NS}AdaptationSet/{NS}Representation"
    )
    if representationElement is None:
        raise ValueError("Representation element not found")

    codecs = representationElement.get("codecs")

    segmentElement = representationElement.find(f"{NS}SegmentTemplate")
    if segmentElement is None:
        raise ValueError("SegmentTemplate element not found")

    url_template = segmentElement.get("media")
    if url_template is None:
        raise ValueError("No `media` attribute in SegmentTemplate")

    timelineElements = segmentElement.findall(f"{NS}SegmentTimeline/{NS}S")
    if not timelineElements:
        raise ValueError("SegmentTimeline elements not found")

    total = 0
    for element in timelineElements:
        total += 1
        count = element.get("r")
        if count is not None:
            total += int(count)

    urls = [url_template.replace("$Number$", str(i)) for i in range(0, total + 1)]

    return urls, codecs


def threadDownload(urls: list[str]) -> bytes:
    # TODO: implement threaded download ⚡️
    # TODO: add progress bar ✨

    data = b""
    for index, url in enumerate(urls):
        req = requests.get(url)
        data += req.content
        print(f"{round((index + 1) / len(urls) * 100)}%")

    return data


def downloadTrackStream(
    path: str, file_name: str, encoded_manifest: str, mime_type: ManifestMimeType
):
    logger.debug(mime_type)
    manifest = decodeManifest(encoded_manifest)

    match mime_type:
        case "application/dash+xml":
            track_urls, codecs = parseManifestXML(manifest)
        case "application/vnd.tidal.bts":
            data = parseManifest(manifest)
            track_urls, codecs = data["urls"], data["codecs"]
        case _:
            raise ValueError(f"Unknown `mime_type`: {mime_type}")

    track_data = threadDownload(track_urls)

    logger.debug(codecs)

    """
    known codecs
        flac (master)
        mp4a.40.2 (high)
        mp4a.40.5 (low)
    """

    makedirs(path, exist_ok=True)

    # TODO: use proper file extension ✨
    file_path = f"{path}/{file_name}.flac"

    with open(file_path, "wb+") as f:
        f.write(track_data)

    return file_path



import time
import logging

from random import randint

from .api import TidalApi
from .auth import getDeviceAuth, getToken, refreshToken
from .config import Config
from .download import downloadTrackStream
from .parser import QUALITY_ARGS, parser
from .types import TRACK_QUALITY, TrackQuality, Track
from .utils import RESOURCE, parseURL, formatFilename


def main():
    args = parser.parse_args()

    logger = logging.getLogger("TIDDL")
    stream_handler = logging.StreamHandler()
    level_name_log = ""
    function_log = ""

    if args.silent:
        log_level = logging.WARNING
    elif args.verbose:
        log_level = logging.DEBUG
        level_name_log = "\033[1;34m%(levelname)s "
        function_log = " %(funcName)s"
    else:
        log_level = logging.INFO

    stream_handler.setLevel(log_level)

    if args.no_color:
        stream_handler.setFormatter(
            logging.Formatter(f"[ %(levelname)s ] %(name)s{function_log} - %(message)s")
        )
    else:
        stream_handler.setFormatter(
            logging.Formatter(
                f"{level_name_log}\033[1;95m%(name)s\033[0;35m{function_log}\033[0m %(message)s"
            )
        )

    file_handler = logging.FileHandler("tiddl.log", "w", "utf-8")
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(
        logging.Formatter("%(levelname)s\t%(name)s.%(funcName)s :: %(message)s")
    )

    logging.basicConfig(
        level=logging.DEBUG,
        handlers=[file_handler, stream_handler],
    )

    config = Config()

    download_path = args.download_path or config["settings"]["download_path"]
    track_quality = (
        QUALITY_ARGS[args.quality]
        if args.quality
        else config["settings"]["track_quality"]
    )

    if args.save_options:
        settings = config.update(
            {
                "settings": {
                    "download_path": download_path,
                    "track_quality": track_quality,
                }
            }
        ).get("settings")

        logger.info(f"saved settings to {config.config_path}")

        if settings:
            print("Current Settings:")
            for k, v in settings.items():
                print(f"\t'{k.upper()}' {v}")

    if not config["token"]:
        auth = getDeviceAuth()
        input(
            f"go to https://{auth['verificationUriComplete']} and add device!\nhit enter when you are ready..."
        )

        # TODO: refresh auth status automatically ✨
        token = getToken(auth["deviceCode"])
        config.update(
            {
                "token": token["access_token"],
                "refresh_token": token["refresh_token"],
                "token_expires_at": int(time.time()) + token["expires_in"],
                "user": {
                    "user_id": str(token["user"]["userId"]),
                    "country_code": token["user"]["countryCode"],
                },
            }
        )
        logger.info(f"authenticated!")

    t_now = int(time.time())
    token_expired = t_now > config["token_expires_at"]

    if token_expired:
        token = refreshToken(config["refresh_token"])
        config.update(
            {
                "token": token["access_token"],
                "token_expires_at": int(time.time()) + token["expires_in"],
            }
        )
        logger.info(f"refreshed token!")

    time_to_expire = config["token_expires_at"] - t_now
    days, hours = time_to_expire // (24 * 3600), time_to_expire % (24 * 3600) // 3600
    days_text = f" {days} {'day' if days == 1 else 'days'}" if days else ""
    hours_text = f" {hours} {'hour' if hours == 1 else 'hours'}" if hours else ""
    logger.info(f"token expires in{days_text}{hours_text}")

    user_input: str = args.input

    if not user_input:
        logger.warning("no ID nor URL provided")
        return

    input_type: RESOURCE
    input_id: str

    if user_input.isdigit():
        input_type = "track"
        input_id = user_input
    else:
        input_type, input_id = parseURL(user_input)

    api = TidalApi(
        config["token"], config["user"]["user_id"], config["user"]["country_code"]
    )

    def downloadTrack(track: Track, sleep=False):
        if sleep:
            sleep_time = randint(10, 30) / 10 + 1
            logger.info(f"sleeping for {sleep_time}s")
            time.sleep(sleep_time)

        file_name = formatFilename(
            args.file_template or config["settings"]["file_template"], track
        )
        stream = api.getTrackStream(track["id"], track_quality)
        quality = TRACK_QUALITY[stream["audioQuality"]]

        MASTER_QUALITIES: list[TrackQuality] = ["HI_RES_LOSSLESS", "LOSSLESS"]
        if stream["audioQuality"] in MASTER_QUALITIES:
            bit_depth, sample_rate = stream.get("bitDepth"), stream.get("sampleRate")
            if bit_depth is None or sample_rate is None:
                raise ValueError(
                    "bitDepth and sampleRate must be provided for master qualities"
                )
            details = f"{bit_depth} bit {sample_rate/1000:.1f} kHz"
        else:
            details = quality["details"]

        logger.info(f"{file_name} :: {quality['name']} Quality - {details}")

        track_path = downloadTrackStream(
            download_path,
            file_name,
            stream["manifest"],
            stream["manifestMimeType"],
        )

        logger.info(f"track saved in {track_path}")

    def downloadAlbum(album_id: str | int):
        # i dont know if limit 100 is suspicious
        # but i will leave it here
        album = api.getAlbumItems(album_id, limit=100)
        for item in album["items"]:
            track = item["item"]
            downloadTrack(track, sleep=True)

    match input_type:
        case "track":
            track = api.getTrack(input_id)
            downloadTrack(track)

        case "album":
            downloadAlbum(input_id)

        case "artist":
            # TODO: include artist EPs and singles option ✨
            artist_albums = api.getArtistAlbums(input_id)

            for album in artist_albums["items"]:
                downloadAlbum(album["id"])

        case "playlist":
            # TODO: add option to limit and set offset of playlist ✨
            playlist = api.getPlaylistItems(input_id)
            for item in playlist["items"]:
                downloadTrack(item["item"], sleep=True)

        case _:
            logger.warning(f"invalid input: `{input_type}`")


if __name__ == "__main__":
    main()



from requests import request
from .types.auth import AuthDeviceResponse, AuthResponse, AuthResponseWithRefresh

AUTH_URL = "https://auth.tidal.com/v1/oauth2"
CLIENT_ID = "zU4XHVVkc2tDPo4t"
CLIENT_SECRET = "VJKhDFqJPqvsPVNBV6ukXTJmwlvbttP7wlMlrc72se4="


def getDeviceAuth() -> AuthDeviceResponse:
    return request(
        "POST",
        f"{AUTH_URL}/device_authorization",
        data={"client_id": CLIENT_ID, "scope": "r_usr+w_usr+w_sub"},
    ).json()


def getToken(device_code: str) -> AuthResponseWithRefresh:
    return request(
        "POST",
        f"{AUTH_URL}/token",
        data={
            "client_id": CLIENT_ID,
            "device_code": device_code,
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            "scope": "r_usr+w_usr+w_sub",
        },
        auth=(CLIENT_ID, CLIENT_SECRET),
    ).json()


def refreshToken(refresh_token: str) -> AuthResponse:
    return request(
        "POST",
        f"{AUTH_URL}/token",
        data={
            "client_id": CLIENT_ID,
            "refresh_token": refresh_token,
            "grant_type": "refresh_token",
            "scope": "r_usr+w_usr+w_sub",
        },
        auth=(CLIENT_ID, CLIENT_SECRET),
    ).json()



from typing import TypedDict, Literal, List, get_args

from .types.track import Track

RESOURCE = Literal["track", "album", "artist", "playlist"]
RESOURCE_LIST: List[RESOURCE] = list(get_args(RESOURCE))


def parseURL(url: str) -> tuple[RESOURCE, str]:
    # remove trailing slash
    url = url.rstrip("/")
    # remove params
    url = url.split("?")[0]

    fragments = url.split("/")

    if len(fragments) < 2:
        raise ValueError(f"Invalid input: {url}")

    parsed_type, parsed_id = fragments[-2], fragments[-1]

    if parsed_type not in RESOURCE_LIST:
        raise ValueError(f"Invalid resource type: {parsed_type} ({url})")

    return parsed_type, parsed_id


class FormattedTrack(TypedDict):
    id: int
    title: str
    number: int
    artist: str
    album: str
    artists: str


def _formatTrackDict(track: Track) -> FormattedTrack:
    artists = [artist["name"] for artist in track["artists"]]
    formatted_track: FormattedTrack = {
        "album": track["album"]["title"],
        "artist": track["artist"]["name"],
        "artists": ", ".join(artists),
        "id": track["id"],
        "title": track["title"],
        "number": track["trackNumber"],
    }
    return formatted_track


def formatFilename(template: str, track: Track) -> str:
    formatted_track = _formatTrackDict(track)
    try:
        return template.format(**formatted_track)
    except KeyError as e:
        missing_key = e.args[0]
        raise ValueError(f"Missing key in track dictionary: {missing_key}")



from setuptools import setup, find_packages

setup(
    name="tiddl",
    version="1.3.0",
    description="TIDDL (Tidal Downloader) is a Python CLI application that allows downloading Tidal tracks.",
    long_description=open('README.md', encoding="utf-8").read(),
    long_description_content_type='text/markdown',
    readme="README.md",
    author="oskvr37",
    packages=find_packages(),
    entry_points={
        "console_scripts": ["tiddl=tiddl:main"],
    },
)
